"""Internal module for I/O related methods.
"""

from __future__ import print_function, absolute_import, division

import codecs
import os


def check_bom(file):
    """Determines file codec from from its BOM record.

    If file starts with BOM record encoded with UTF-8 or UTF-16(BE/LE)
    then corresponding encoding name is returned, otherwise None is returned.
    In both cases file current position is set to after-BOM bytes. The file
    must be open in binary mode and positioned at offset 0.
    """

    # try to read first three bytes
    lead = file.read(3)
    if len(lead) == 3 and lead == codecs.BOM_UTF8:
        # UTF-8, position is already OK, use canonical name
        return codecs.lookup('utf-8').name
    elif len(lead) >= 2 and lead[:2] == codecs.BOM_UTF16_BE:
        # need to backup one character
        if len(lead) == 3:
            file.seek(-1, os.SEEK_CUR)
        return codecs.lookup('utf-16-be').name
    elif len(lead) >= 2 and lead[:2] == codecs.BOM_UTF16_LE:
        # need to backup one character
        if len(lead) == 3:
            file.seek(-1, os.SEEK_CUR)
        return codecs.lookup('utf-16-le').name
    else:
        # no BOM, rewind
        file.seek(-len(lead), os.SEEK_CUR)
        return None


def guess_lineno(file):
    """Guess current line number in a file.

    Guessing is done in a very crude way - scanning file from beginning
    until current offset and counting newlines. Only meant to be used in
    exceptional cases - generating line number for error message.
    """
    offset = file.tell()
    file.seek(0)
    startpos = 0
    lineno = 1
    # looks like file.read() return bytes in python3
    # so I need more complicated algorithm here
    while True:
        line = file.readline()
        if not line:
            break
        endpos = file.tell()
        if startpos <= offset < endpos:
            break
        lineno += 1
    file.seek(offset)
    return lineno



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ged4py.detail.name` module."""

import unittest

from ged4py.detail.name import (split_name, parse_name_altree,
                                parse_name_myher, parse_name_ancestris)
from ged4py import model


class TestDetailName(unittest.TestCase):
    """Tests for `ged4py.detail.io` module."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_001_split_name(self):
        """Test detail.name.split_name()."""

        nsplit = split_name("First Name Only")
        self.assertEqual(nsplit, ("First Name Only", "", ""))

        nsplit = split_name("First Name /Last Name/")
        self.assertEqual(nsplit, ("First Name", "Last Name", ""))

        nsplit = split_name("First/ Last / ")
        self.assertEqual(nsplit, ("First", "Last", ""))

        nsplit = split_name(" First /Last ")
        self.assertEqual(nsplit, ("First", "Last", ""))

        nsplit = split_name("First /Last/ II ")
        self.assertEqual(nsplit, ("First", "Last", "II"))

        nsplit = split_name("/Last/ Karl II")
        self.assertEqual(nsplit, ("", "Last", "Karl II"))

        nsplit = split_name("Жанна /Иванова (Д'Арк)/")
        self.assertEqual(nsplit, ("Жанна", "Иванова (Д'Арк)", ""))

    def test_002_parse_name_altree(self):
        """Test parse_name_altree()
        """
        rec = model.NameRec()
        rec.level = 1
        rec.tag = "NAME"
        rec.dialect = model.DIALECT_ALTREE

        rec.value = "First /Last Name/ Second"
        name_tup = parse_name_altree(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last Name", "Second"))

        rec.value = "First /Last(-er)/ Second"
        name_tup = parse_name_altree(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last(-er)", "Second"))

        rec.value = "First /Last (-er)/ Second"
        name_tup = parse_name_altree(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last (-er)", "Second"))

        maiden = model.Record()
        maiden.level = 2
        maiden.tag = "SURN"
        maiden.value = "Maiden"
        rec.value = "First /Last (Maiden)/"
        rec.sub_records = [maiden]
        name_tup = parse_name_altree(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last", "", "Maiden"))

    def test_003_parse_name_myher(self):
        """Test parse_name_myher()
        """
        rec = model.NameRec()
        rec.level = 1
        rec.tag = "NAME"
        rec.dialect = model.DIALECT_MYHERITAGE

        rec.value = "First /Last Name/ Second"
        name_tup = parse_name_myher(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last Name", "Second"))

        rec.value = "First /Last(-er)/ Second"
        name_tup = parse_name_myher(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last(-er)", "Second"))

        rec.value = "First /Last (-er)/ Second"
        name_tup = parse_name_altree(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last (-er)", "Second"))

        married = model.Record()
        married.level = 2
        married.tag = "_MARNM"
        married.value = "Married"
        rec.value = "First /Maiden/"
        rec.sub_records = [married]
        name_tup = parse_name_myher(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Married", "", "Maiden"))

    def test_004_parse_name_ancestris(self):
        """Test parse_name_ancestris()

        Ancestris dialect is just a split_name() which is tested separately,
        so the is not much testing needed for it.
        """
        rec = model.NameRec()
        rec.level = 1
        rec.tag = "NAME"
        rec.dialect = model.DIALECT_ANCESTRIS

        rec.value = "First /Last Name/ Second"
        name_tup = parse_name_ancestris(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last Name", "Second"))

        rec.value = "First /Last(-er)/ Second"
        name_tup = parse_name_ancestris(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last(-er)", "Second"))

        rec.value = "First /Last (-er)/ Second"
        name_tup = parse_name_ancestris(rec)
        self.assertIsInstance(name_tup, tuple)
        self.assertEqual(name_tup, ("First", "Last (-er)", "Second"))



"""Internal module for parsing names in gedcom format.
"""

from __future__ import print_function, absolute_import, division


def split_name(name):
    """Extracts pieces of name from full name string.

    Full name can have one of these formats:
        <NAME_TEXT> |
        /<NAME_TEXT>/ |
        <NAME_TEXT> /<NAME_TEXT>/ |
        /<NAME_TEXT>/ <NAME_TEXT> |
        <NAME_TEXT> /<NAME_TEXT>/ <NAME_TEXT>

    <NAME_TEXT> can include almost anything excluding commas, numbers,
    special characters (though some test files use numbers for the names).
    Text between slashes is considered a surname, outside slashes - given
    name.

    This method splits full name into pieces at slashes, e.g.:

        "First /Last/" -> ("First", "Last", "")
        "/Last/ First" -> ("", "Last", "First")
        "First /Last/ Jr." -> ("First", "Last", "Jr.")
        "First Jr." -> ("First Jr.", "", "")

    :param str name: Full name string.
    :return: 2-tuple `(given1, surname, given2)`, `surname` or `given` will
        be empty strings if they are not present in full string.
    """
    given1, _, rem = name.partition("/")
    surname, _, given2 = rem.partition("/")
    return given1.strip(), surname.strip(), given2.strip()


def parse_name_altree(record):
    """Parse NAME structure assuming ALTREE dialect.

    In ALTREE dialect maiden name (if present) is saved as SURN sub-record
    and is also appended to family name in parens. Given name is saved in
    GIVN sub-record. Few examples:

    No maiden name:

        1 NAME John /Smith/
        2 GIVN John

    With maiden name:

        1 NAME Jane /Smith (Ivanova)/
        2 GIVN Jane
        2 SURN Ivanova

    No maiden name

        1 NAME Mers /Daimler (-Benz)/
        2 GIVN Mers

    Because family name can also contain parens it's not enough to parse
    family name and guess maiden name from it, we also have to check for
    SURN record.

    ALTREE also replaces empty names with question mark, we undo that too.

    :param record: NAME record
    :return: tuple with 3 or 4 elements, first three elements of tuple are
        the same as returned from :py:meth:`split_name` method, fourth element
        (if present) denotes maiden name.
    """
    name_tuple = split_name(record.value)
    if name_tuple[1] == '?':
        name_tuple = (name_tuple[0], '', name_tuple[2])
    maiden = record.sub_tag_value("SURN")
    if maiden:
        # strip "(maiden)" from family name
        ending = '(' + maiden + ')'
        surname = name_tuple[1]
        if surname.endswith(ending):
            surname = surname[:-len(ending)].rstrip()
            if surname == '?':
                surname = ''
        name_tuple = (name_tuple[0], surname, name_tuple[2], maiden)
    return name_tuple


def parse_name_myher(record):
    """Parse NAME structure assuming MYHERITAGE dialect.

    In MYHERITAGE dialect married name (if present) is saved as _MARNM
    sub-record. Maiden name is stored in SURN record. Few examples:

    No maiden name:

        1 NAME John /Smith/
        2 GIVN John
        2 SURN Smith

    With maiden name:

        1 NAME Jane /Ivanova/
        2 GIVN Jane
        2 SURN Ivanova
        2 _MARNM Smith

    No maiden name

        1 NAME Mers /Daimler (-Benz)/
        2 GIVN Mers
        2 SURN Daimler (-Benz)


    :param record: NAME record
    :return: tuple with 3 or 4 elements, first three elements of tuple are
        the same as returned from :py:meth:`split_name` method, fourth element
        (if present) denotes maiden name.
    """
    name_tuple = split_name(record.value)
    married = record.sub_tag_value("_MARNM")
    if married:
        maiden = name_tuple[1]
        name_tuple = (name_tuple[0],
                      married,
                      name_tuple[2],
                      maiden)
    return name_tuple


def parse_name_ancestris(record):
    """Parse NAME structure assuming ANCESTRIS dialect.

    As far as I can tell Ancestris does not have any standard convention for
    representing maiden or married names. Best we can do in this situation is
    to use NAME record value and ignore any other fields.

    :param record: NAME record
    :return: tuple with 3 or 4 elements, first three elements of tuple are
        the same as returned from :py:meth:`split_name` method, fourth element
        (if present) denotes maiden name.
    """
    name_tuple = split_name(record.value)
    return name_tuple



"""Internal module for parsing dates in gedcom format.
"""

from __future__ import print_function, absolute_import, division

__all__ = ["CalendarDate", "DateValue"]

import re
import string


MONTHS_GREG = ['JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN', 'JUL', 'AUG',
               'SEP', 'OCT', 'NOV', 'DEC']
MONTHS_HEBR = ['TSH', 'CSH', 'KSL', 'TVT', 'SHV', 'ADR', 'ADS', 'NSN',
               'IYR', 'SVN', 'TMZ', 'AAV', 'ELL']
MONTHS_FREN = ['VEND', 'BRUM', 'FRIM', 'NIVO', 'PLUV', 'VENT', 'GERM',
               'FLOR', 'PRAI', 'MESS', 'THER', 'FRUC', 'COMP']

# DATE_VALUE := [
#     <DATE> |
#     <DATE_PERIOD> |
#     <DATE_RANGE>|
#     <DATE_APPROXIMATED> |
#     INT <DATE> (<DATE_PHRASE>) |
#     (<DATE_PHRASE>)
#     ]

# DATE := [<DATE_CALENDAR_ESCAPE> | <NULL>] <DATE_CALENDAR>
# <DATE_CALENDAR> := [<YEAR> | <MONTH> <YEAR> | <DAY> <MONTH> <YEAR>]
# <YEAR can be specified as "1000B.C." or "1699/00"
# <MONTH> is all characters.
# This does not use named groups, it may appear few times in other expressions
# Groups: 1: calendar; 2: day; 3: month; 4: year
DATE = r"""
    (?:@\#D(\w+)@\s+)?          # @#DCALENDAR@, optional
    (?:
        (?:(\d+)\s+)?           # day (int), optional
        ([A-Z]{3,4})\s+         # month, name 3-4 chars,
    )?
    (\d+\S*)                    # year, required, number with optional suffix
    """
DATE_RE = re.compile("^" + DATE + "$", re.X | re.I)

# DATE_PERIOD:= [ FROM <DATE> | TO <DATE> | FROM <DATE> TO <DATE> ]
DATE_PERIOD_FROM = r"^FROM\s+(?P<date>" + DATE + ")$"
DATE_PERIOD_TO = r"^TO\s+(?P<date>" + DATE + ")$"
DATE_PERIOD = r"^FROM\s+(?P<date1>" + DATE + r")\s+TO\s+(?P<date2>" + \
              DATE + ")$"

# DATE_RANGE:= [ BEF <DATE> | AFT <DATE> | BET <DATE> AND <DATE> ]
DATE_RANGE_BEFORE = r"^BEF\s+(?P<date>" + DATE + ")$"
DATE_RANGE_AFTER = r"^AFT\s+(?P<date>" + DATE + ")$"
DATE_RANGE = r"^BET\s+(?P<date1>" + DATE + r")\s+AND\s+(?P<date2>" + \
             DATE + ")$"

# DATE_APPROXIMATED := [ ABT <DATE> | CAL <DATE> | EST <DATE> ]
DATE_APPROX_ABOUT = r"^ABT\s+(?P<date>" + DATE + ")$"
DATE_APPROX_CALC = r"^CAL\s+(?P<date>" + DATE + ")$"
DATE_APPROX_EST = r"^EST\s+(?P<date>" + DATE + ")$"

# INT <DATE> (<DATE_PHRASE>)
DATE_INTERP = r"^INT\s+(?P<date>" + DATE + r")\s+\((?P<phrase>.*)\)$"
DATE_PHRASE = r"^\((?P<phrase>.*)\)$"

# INT <DATE> (<DATE_PHRASE>)
DATE_SIMPLE = r"^(?P<date>" + DATE + ")$"

DATES = ((re.compile(DATE_PERIOD, re.X | re.I), "FROM $date1 TO $date2"),
         (re.compile(DATE_PERIOD_FROM, re.X | re.I), "FROM $date"),
         (re.compile(DATE_PERIOD_TO, re.X | re.I), "TO $date"),
         (re.compile(DATE_RANGE, re.X | re.I), "BETWEEN $date1 AND $date2"),
         (re.compile(DATE_RANGE_BEFORE, re.X | re.I), "BEFORE $date"),
         (re.compile(DATE_RANGE_AFTER, re.X | re.I), "AFTER $date"),
         (re.compile(DATE_APPROX_ABOUT, re.X | re.I), "ABOUT $date"),
         (re.compile(DATE_APPROX_CALC, re.X | re.I), "CALCULATED $date"),
         (re.compile(DATE_APPROX_EST, re.X | re.I), "ESTIMATED $date"),
         (re.compile(DATE_INTERP, re.X | re.I), "INTERPRETED $date ($phrase)"),
         (re.compile(DATE_PHRASE, re.X | re.I), "($phrase)"),
         (re.compile(DATE_SIMPLE, re.X | re.I), "$date"),
         )


class CalendarDate(object):
    """Representation of calendar date, corresponding to <DATE> element.

    This includes optional calendar kind (calendar escape) which can be
    one of @#DHEBREW@ | @#DROMAN@ | @#DFRENCH R@ | @#DGREGORIAN@ |
    @#DJULIAN@ | @#DUNKNOWN@ (@#DGREGORIAN@) is default. Date consists
    of year, month, and day; day and moth are optional (eitehr day or
    day+month), year must be present. Day is a number, month is month name
    in given calendar. Year is a number optionally followed by "B.C." or
    "/NN".

    Dates can be ordered but ordering only works for the dates from the
    same calendar, dates from different calendars will be ordered randomly.

    In general date parsing tries to make best effort to guess all possible
    formats, but if it fails it does not raise exceptions, instead it uses
    some defaults values for items that cannot be understood.

    :param str year: String representing year in a calendar. Expected to
        start with few digits and followed by some suffix.
    :param str month: Name of the month. Optional, but if day is given then
        month cannot be None.
    :param int day: Day in a month, optional.
    :param str calendar: one of "GREGORIAN", "JULIAN", "HEBREW", "ROMAN",
        "FRENCH R", "UNKNOWN", default is "GREGORIAN"
    """

    DIGITS = re.compile(r"\d+")
    MONTHS = {"GREGORIAN": MONTHS_GREG,
              "JULIAN": MONTHS_GREG,
              "HEBREW": MONTHS_HEBR,
              "FRENCH R": MONTHS_FREN}

    def __init__(self, year=None, month=None, day=None, calendar=None):
        self.year = year
        self.month = None if month is None else month.upper()
        self.day = day
        self.calendar = calendar or "GREGORIAN"

        # determine month number
        months = self.MONTHS.get(self.calendar, [])
        try:
            self.month_num = months.index(self.month) + 1
        except ValueError:
            self.month_num = None

        self._tuple = None

    @classmethod
    def parse(cls, datestr):
        """Parse <DATE> string and make :py:class:`CalendarDate` from it.

        :param str datestr: String with GEDCOM date.
        """
        m = DATE_RE.match(datestr)
        if m is not None:
            day = None if m.group(2) is None else int(m.group(2))
            return cls(m.group(4), m.group(3), day, m.group(1))

    @property
    def as_tuple(self):
        """Date as three-tuple of numbers"""
        if self._tuple is None:
            # extract leading digits from year
            year = 9999
            if self.year:
                m = self.DIGITS.match(self.year)
                if m:
                    year = int(m.group(0))
            month = self.month_num or 99
            day = self.day if self.day is not None else 99

            # should we include calendar name in tuple too?
            self._tuple = year, month, day

        return self._tuple

    def __lt__(self, other):
        return self.as_tuple < other.as_tuple

    def __le__(self, other):
        return self.as_tuple <= other.as_tuple

    def __eq__(self, other):
        return self.as_tuple == other.as_tuple

    def __ne__(self, other):
        return self.as_tuple != other.as_tuple

    def __gt__(self, other):
        return self.as_tuple > other.as_tuple

    def __ge__(self, other):
        return self.as_tuple >= other.as_tuple

    def fmt(self):
        """Make printable representation out of this instance.
        """
        val = str(self.year)
        if self.month is not None:
            val += ' ' + str(self.month)
            if self.day is not None:
                val += ' ' + str(self.day)
        return val

    def __str__(self):
        return "{}(year={}, month={}, day={}, calendar={})".format(
            self.__class__.__name__, self.year, self.month,
            self.day, self.calendar)

    def __repr__(self):
        return str(self)


class DateValue(object):
    """Representation of the <DATE_VALUE>, can be exact date, range,
    period, etc.

    If `kw` is empty (default) then resulting DateValue is guaranteed to be
    in the future w.r.t. any regular dates.

    :param str tmpl: Template string acceptable by `string.Template`.
    :param dict kw: Dictionary with the keys being keywords in template
        string and values are instance of :py:class:`CalendarDate` or
        strings.
    """

    def __init__(self, tmpl="", kw={}):
        self.template = tmpl
        self.kw = kw

    @classmethod
    def parse(cls, datestr):
        """Parse string <DATE_VALUE> string and make :py:class:`DateValue`
        instance out of it.

        :param str datestr: String with GEDCOM date, range, period, etc.
        """
        # some apps generate DATE recods without any value, which is
        # non-standard, return empty DateValue for those
        if not datestr:
            return cls()
        for regex, tmpl in DATES:
            m = regex.match(datestr)
            if m is not None:
                groups = {}
                for key, val in m.groupdict().items():
                    if key != 'phrase':
                        val = CalendarDate.parse(val)
                    groups[key] = val
                return cls(tmpl, groups)
        # if cannot parse string assume it is a phrase
        return cls("($phrase)", dict(phrase=datestr))

    @property
    def _cmp_date(self):
        """Returns Calendar date used for comparison.

        Use the earliest date out of all CalendarDates in this instance,
        or some date in the future if there are no CalendarDates (e.g.
        when Date is a phrase).
        """
        dates = sorted(val for val in self.kw.values()
                       if isinstance(val, CalendarDate))
        if dates:
            return dates[0]
        # return date very far in the future
        return CalendarDate()

    def __lt__(self, other):
        return self._cmp_date < other._cmp_date

    def __le__(self, other):
        return self._cmp_date <= other._cmp_date

    def __eq__(self, other):
        return self._cmp_date == other._cmp_date

    def __ne__(self, other):
        return self._cmp_date != other._cmp_date

    def __gt__(self, other):
        return self._cmp_date > other._cmp_date

    def __ge__(self, other):
        return self._cmp_date >= other._cmp_date

    def fmt(self):
        """Make printable representation out of this instance.
        """
        tmpl = string.Template(self.template)
        kw = {}
        for key, val in self.kw.items():
            if key == 'phrase':
                kw[key] = val
            else:
                kw[key] = val.fmt()
        return tmpl.substitute(kw)

    def __str__(self):
        return "{}({})".format(self.__class__.__name__, self.fmt())

    def __repr__(self):
        return str(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for ged4py encodings handling."""

import io
import logging
import pytest

from ged4py.parser import GedcomReader, CodecError


def _check_log_rec(rec, level, msg, args):
    assert rec.levelno == level
    assert msg in rec.msg
    assert rec.args == args


def test_001_standard():
    """Test standard encodings."""

    file = io.BytesIO(b"0 HEAD\n1 CHAR ASCII\n0 TRLR")
    reader = GedcomReader(file)
    assert reader._encoding == "ascii"

    file = io.BytesIO(b"0 HEAD\n1 CHAR ANSEL\n0 TRLR")
    reader = GedcomReader(file)
    assert reader._encoding == "gedcom"

    file = io.BytesIO(b"0 HEAD\n1 CHAR UTF-8\n0 TRLR")
    reader = GedcomReader(file)
    assert reader._encoding == "utf-8"

    file = io.BytesIO(b"\xef\xbb\xbf0 HEAD\n1 CHAR UTF-8\n0 TRLR")
    reader = GedcomReader(file)
    assert reader._encoding == "utf-8"

    # UTF-16 is broken, do not use


@pytest.mark.parametrize('enc,pyenc,ambig',
                         [("IBMPC", "cp437", True),
                          ("IBM", "cp437", True),
                          ("IBM-PC", "cp437", True),
                          ("OEM", "cp437", True),
                          ("MSDOS", "cp850", True),
                          ("IBM DOS", "cp850", True),
                          ("MS-DOS", "cp850", True),
                          ("ANSI", "cp1252", True),
                          ("WINDOWS", "cp1252", True),
                          ("IBM WINDOWS", "cp1252", True),
                          ("IBM_WINDOWS", "cp1252", True),
                          ("WINDOWS-1250", "cp1250", False),
                          ("WINDOWS-1251", "cp1251", False),
                          ("CP1252", "cp1252", False),
                          ("ISO-8859-1", "iso8859-1", False),
                          ("ISO8859-1", "iso8859-1", False),
                          ("ISO8859", "iso8859-1", True),
                          ("LATIN1", "iso8859-1", True),
                          ("MACINTOSH", "mac-roman", True),
                          ])
def test_002_illegal(enc, pyenc, ambig, caplog):
    """Test for illegal encodings.
    """
    caplog.set_level(logging.WARNING)

    # %s formatting works in py27 and py3
    char = ("1 CHAR " + enc).encode()
    file = io.BytesIO(b"0 HEAD\n" + char + b"\n0 TRLR")
    reader = GedcomReader(file)
    assert reader._encoding == pyenc

    # check logging
    assert len(caplog.records) == (2 if ambig else 1)
    _check_log_rec(caplog.records[0], logging.ERROR,
                   "is not a legal character set or encoding",
                   (2, char, enc))
    if ambig:
        _check_log_rec(caplog.records[1], logging.WARNING,
                       "is ambiguous, it will be interpreted as",
                       (enc, pyenc))


def test_003_codec_exceptions():
    """Test codecs-related exceptions."""

    # unknown codec name
    file = io.BytesIO(b"0 HEAD\n1 CHAR NOTCODEC\n0 TRLR")
    with pytest.raises(CodecError):
        GedcomReader(file)

    # BOM disagrees with CHAR
    file = io.BytesIO(b"\xef\xbb\xbf0 HEAD\n1 CHAR ANSEL\n0 TRLR")
    with pytest.raises(CodecError):
        GedcomReader(file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ged4py.detail.io` module."""

from contextlib import contextmanager
import io
import tempfile
import os
import unittest

from ged4py.detail.io import check_bom, guess_lineno


@contextmanager
def _temp_file(data):
    """Create file with unique name and store some data in it.

    Returns file name.
    """
    fd, fname = tempfile.mkstemp()
    os.write(fd, data)
    os.close(fd)
    yield fname
    os.unlink(fname)


class TestDetailIo(unittest.TestCase):
    """Tests for `ged4py.detail.io` module."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_001_guess_bom_codec(self):
        """Test detail.io.check_bom()."""

        file = io.BytesIO(b"0 HEAD")
        codec = check_bom(file)
        self.assertTrue(codec is None)
        self.assertEqual(file.tell(), 0)

        file = io.BytesIO(b"0")
        codec = check_bom(file)
        self.assertTrue(codec is None)
        self.assertEqual(file.tell(), 0)

        file = io.BytesIO(b"\xef\xbb\xbf0 HEAD")
        codec = check_bom(file)
        self.assertEqual(codec, "utf-8")
        self.assertEqual(file.tell(), 3)

        file = io.BytesIO(b"\xff\xfe0 HEAD")
        codec = check_bom(file)
        self.assertEqual(codec, "utf-16-le")
        self.assertEqual(file.tell(), 2)

        file = io.BytesIO(b"\xfe\xff0 HEAD")
        codec = check_bom(file)
        self.assertEqual(codec, "utf-16-be")
        self.assertEqual(file.tell(), 2)

        file = io.BytesIO(b"\xfe\xff")
        codec = check_bom(file)
        self.assertEqual(codec, "utf-16-be")
        self.assertEqual(file.tell(), 2)

    def test_002_guess_lineno(self):
        """Test detail.io.guess_lineno()."""

        file = io.BytesIO(b"line1\nline2\nline3\nline4\nline5\n")
        file.readline()
        self.assertEqual(file.tell(), 6)
        self.assertEqual(guess_lineno(file), 2)
        self.assertEqual(file.tell(), 6)

        file.readline()
        self.assertEqual(file.tell(), 12)
        self.assertEqual(guess_lineno(file), 3)
        self.assertEqual(file.tell(), 12)

        file.readline()
        self.assertEqual(file.tell(), 18)
        self.assertEqual(guess_lineno(file), 4)
        self.assertEqual(file.tell(), 18)

        file.readline()
        self.assertEqual(file.tell(), 24)
        self.assertEqual(guess_lineno(file), 5)
        self.assertEqual(file.tell(), 24)

        file.readline()
        self.assertEqual(file.tell(), 30)
        self.assertEqual(guess_lineno(file), 6)
        self.assertEqual(file.tell(), 30)

        file.readline()
        self.assertEqual(file.tell(), 30)
        self.assertEqual(guess_lineno(file), 6)
        self.assertEqual(file.tell(), 30)

        file.seek(0)
        self.assertEqual(file.tell(), 0)
        self.assertEqual(guess_lineno(file), 1)
        self.assertEqual(file.tell(), 0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ged4py.model` module."""

import unittest

from ged4py import model
from ged4py.detail.date import DateValue


class TestModel(unittest.TestCase):
    """Tests for `ged4py.model` module."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_001_record(self):
        """Test Record class."""

        rec = model.Record()
        self.assertEqual(vars(rec), {'dialect': None,
                                     'xref_id': None,
                                     'level': None,
                                     'value': None,
                                     'tag': None,
                                     'sub_records': None,
                                     'offset': None})

        rec.level = 0
        rec.xref_id = "@x@"
        rec.tag = "REC"
        rec.value = "value"
        rec.sub_records = []
        rec.offset = 1000
        rec.dialect = model.DIALECT_DEFAULT

        rec.freeze()
        self.assertTrue(rec.sub_tag("SUB") is None)
        self.assertEqual(rec.sub_tags("SUB"), [])

    def test_002_record_sub(self):
        """Test Record class with sub-records."""

        rec = model.Record()
        rec.level = 0
        rec.xref_id = "@x@"
        rec.tag = "REC"
        rec.value = "value"
        rec.sub_records = []
        rec.offset = 1000
        rec.dialect = model.DIALECT_DEFAULT

        for subtag in ['SUBA', 'SUBB', 'SUBC']:
            for i in range(3):

                sub = model.Record()
                sub.level = 1
                sub.tag = subtag
                sub.value = i
                sub.sub_records = []
                sub.offset = 1100 + 100 * i
                sub.dialect = model.DIALECT_DEFAULT

                subsub = model.Record()
                subsub.level = 2
                subsub.tag = "SUB"
                subsub.value = 'VALUE'
                subsub.sub_records = []
                subsub.dialect = model.DIALECT_DEFAULT
                subsub.freeze()

                sub.sub_records.append(subsub)
                sub.freeze()

                rec.sub_records.append(sub)

        # direct sub-tags
        rec.freeze()
        for subtag in ['SUBA', 'SUBB', 'SUBC']:
            self.assertEqual(rec.sub_tag(subtag).tag, subtag)
            self.assertEqual(len(rec.sub_tags(subtag)), 3)

        # non-existing sub-tags
        self.assertTrue(rec.sub_tag("SUB") is None)
        self.assertEqual(rec.sub_tags("SUB"), [])
        self.assertTrue(rec.sub_tag_value("SUB") is None)

        # hierarchical sub-tags
        self.assertTrue(rec.sub_tag("SUBA/SUB") is not None)
        self.assertEqual(rec.sub_tag("SUBB/SUB").tag, "SUB")
        self.assertEqual(rec.sub_tag_value("SUBB/SUB"), "VALUE")

    def test_010_namerec_default(self):
        """Test NameRec class with default dialect."""

        rec = model.NameRec()
        rec.level = 1
        rec.tag = "NAME"
        rec.dialect = model.DIALECT_DEFAULT

        rec.value = ""
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "", ""))

        rec.value = None
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "", ""))

        rec.value = "First Name"
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First Name", "", ""))

        rec.value = " /Last Name/ "
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "Last Name", ""))

        rec.value = " /Last Name"
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "Last Name", ""))

        rec.value = "First /Last Name/ "
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First", "Last Name", ""))

        rec.value = "First /Last Name/ Second "
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First", "Last Name", "Second"))

        rec.value = "First /Last (Maiden)/"
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First", "Last (Maiden)", ""))

    def test_011_namerec_altree(self):
        """Test NameRec class with ALTREE dialect."""

        rec = model.NameRec()
        rec.level = 1
        rec.tag = "NAME"
        rec.dialect = model.DIALECT_ALTREE

        rec.value = ""
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "", ""))

        rec.value = None
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "", ""))

        rec.value = "First /Last Name/ Second "
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First", "Last Name", "Second"))

        maiden = model.Record()
        maiden.level = 2
        maiden.tag = "SURN"
        maiden.value = "Maiden"
        rec.value = "First /Last (Maiden)/"
        rec.sub_records = [maiden]
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First", "Last", "", "Maiden"))

    def test_012_namerec_myher(self):
        """Test NameRec class with MYHERITAGE dialect."""

        rec = model.NameRec()
        rec.level = 1
        rec.tag = "NAME"
        rec.dialect = model.DIALECT_MYHERITAGE
        rec.sub_records = []

        rec.value = ""
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "", ""))

        rec.value = None
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("", "", ""))

        rec.value = "First /Last Name/ Second "
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First", "Last Name", "Second"))

        married = model.Record()
        married.level = 2
        married.tag = "_MARNM"
        married.value = "Last"
        married.sub_records = []
        rec.value = "First /Maiden/"
        rec.sub_records = [married]
        rec.freeze()
        self.assertIsInstance(rec.value, tuple)
        self.assertEqual(rec.value, ("First", "Last", "", "Maiden"))

    def test_020_name_default(self):
        """Test Name class with DEFAULT dialect."""

        dialect = model.DIALECT_DEFAULT
        names = [model.make_record(1, None, "NAME", "John /Smith/", [], 0, dialect).freeze()]
        name = model.Name(names, dialect)

        self.assertTrue(name._primary is names[0])
        self.assertEqual(name.surname, "Smith")
        self.assertEqual(name.given, "John")
        self.assertTrue(name.maiden is None)

        self.assertEqual(name.order(model.ORDER_SURNAME_GIVEN), ("1Smith", "1John"))
        self.assertEqual(name.order(model.ORDER_GIVEN_SURNAME), ("1John", "1Smith"))
        self.assertEqual(name.order(model.ORDER_MAIDEN_GIVEN), ("1Smith", "1John"))
        self.assertEqual(name.order(model.ORDER_GIVEN_MAIDEN), ("1John", "1Smith"))

        self.assertEqual(name.format(), ("John Smith"))

        names = [model.make_record(1, None, "NAME", "John", [], 0, dialect).freeze()]
        name = model.Name(names, dialect)

        self.assertTrue(name._primary is names[0])
        self.assertEqual(name.surname, "")
        self.assertEqual(name.given, "John")
        self.assertTrue(name.maiden is None)

        self.assertEqual(name.order(model.ORDER_SURNAME_GIVEN), ("2", "1John"))
        self.assertEqual(name.order(model.ORDER_GIVEN_SURNAME), ("1John", "2"))
        self.assertEqual(name.order(model.ORDER_MAIDEN_GIVEN), ("2", "1John"))
        self.assertEqual(name.order(model.ORDER_GIVEN_MAIDEN), ("1John", "2"))

        self.assertEqual(name.format(), ("John"))

        name_type = model.make_record(2, None, "TYPE", "maiden", [], 0, dialect).freeze()
        names = [model.make_record(1, None, "NAME", "/Sawyer/", [name_type], 0, dialect).freeze(),
                 model.make_record(1, None, "NAME", "Jane /Smith/ A.", [], 0, dialect).freeze()]
        name = model.Name(names, dialect)

        self.assertTrue(name._primary is names[1])
        self.assertEqual(name.surname, "Smith")
        self.assertEqual(name.given, "Jane A.")
        self.assertEqual(name.maiden, "Sawyer")

        self.assertEqual(name.order(model.ORDER_SURNAME_GIVEN), ("1Smith", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_SURNAME), ("1Jane A.", "1Smith"))
        self.assertEqual(name.order(model.ORDER_MAIDEN_GIVEN), ("1Sawyer", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_MAIDEN), ("1Jane A.", "1Sawyer"))

        self.assertEqual(name.format(), ("Jane Smith A."))

    def test_021_name_altree(self):
        """Test Name class with ALTREE dialect."""

        dialect = model.DIALECT_ALTREE
        surn = model.make_record(2, None, "SURN", "Sawyer", [], 0, dialect).freeze()
        names = [model.make_record(1, None, "NAME", "Jane /Smith (Sawyer)/ A.", [surn], 0, dialect).freeze()]
        name = model.Name(names, dialect)

        self.assertTrue(name._primary is names[0])
        self.assertEqual(name.surname, "Smith")
        self.assertEqual(name.given, "Jane A.")
        self.assertEqual(name.maiden, "Sawyer")

        self.assertEqual(name.order(model.ORDER_SURNAME_GIVEN), ("1Smith", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_SURNAME), ("1Jane A.", "1Smith"))
        self.assertEqual(name.order(model.ORDER_MAIDEN_GIVEN), ("1Sawyer", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_MAIDEN), ("1Jane A.", "1Sawyer"))

        self.assertEqual(name.format(), ("Jane Smith A."))

        names = [model.make_record(1, None, "NAME", "Jane /?/ A.", [], 0, dialect).freeze()]
        name = model.Name(names, dialect)

        self.assertTrue(name._primary is names[0])
        self.assertEqual(name.surname, "")
        self.assertEqual(name.given, "Jane A.")
        self.assertTrue(name.maiden is None)

        self.assertEqual(name.order(model.ORDER_SURNAME_GIVEN), ("2", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_SURNAME), ("1Jane A.", "2"))
        self.assertEqual(name.order(model.ORDER_MAIDEN_GIVEN), ("2", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_MAIDEN), ("1Jane A.", "2"))

        self.assertEqual(name.format(), ("Jane A."))

    def test_022_name_myher(self):
        """Test Name class with MYHERITAGE dialect."""

        dialect = model.DIALECT_MYHERITAGE
        married = model.make_record(2, None, "_MARNM", "Smith", [], 0, dialect).freeze()
        names = [model.make_record(1, None, "NAME", "Jane /Sawyer/ A.", [married], 0, dialect).freeze()]
        name = model.Name(names, dialect)

        self.assertTrue(name._primary is names[0])
        self.assertEqual(name.surname, "Smith")
        self.assertEqual(name.given, "Jane A.")
        self.assertEqual(name.maiden, "Sawyer")

        self.assertEqual(name.order(model.ORDER_SURNAME_GIVEN), ("1Smith", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_SURNAME), ("1Jane A.", "1Smith"))
        self.assertEqual(name.order(model.ORDER_MAIDEN_GIVEN), ("1Sawyer", "1Jane A."))
        self.assertEqual(name.order(model.ORDER_GIVEN_MAIDEN), ("1Jane A.", "1Sawyer"))

        self.assertEqual(name.format(), ("Jane Smith A."))

    def test_030_date(self):
        """Test Date class."""

        dialect = model.DIALECT_MYHERITAGE
        date = model.make_record(1, None, "DATE", "1970", [], 0, dialect).freeze()
        self.assertIsInstance(date, model.Date)
        self.assertIsInstance(date.value, DateValue)

        # empty date tag makes Date with empty DateValue
        dialect = model.DIALECT_DEFAULT
        date = model.make_record(1, None, "DATE", None, [], 0, dialect).freeze()
        self.assertIsInstance(date, model.Date)
        self.assertIsInstance(date.value, DateValue)
        self.assertEqual(date.value.template, "")

        date = model.make_record(1, None, "DATE", "", [], 0, dialect).freeze()
        self.assertIsInstance(date, model.Date)
        self.assertIsInstance(date.value, DateValue)
        self.assertEqual(date.value.template, "")

    def test_040_Pointer(self):
        """Test Pointer class."""

        class Parser(object):
            def __init__(self):
                self.xref0 = {b"@pointer0@": (0, "TAG0"),
                              b"@pointer1@": (1, "TAG1")}

            def read_record(self, offset):
                return str(offset)

        pointer = model.Pointer(Parser())
        pointer.value = b"@pointer0@"
        pointer.freeze()

        self.assertEqual(pointer.value, b"@pointer0@")
        self.assertEqual(pointer.ref, "0")

        dialect = model.DIALECT_MYHERITAGE
        pointer = model.make_record(1, None, "FAMC", b"@pointer1@", [], 0, dialect, Parser()).freeze()
        self.assertIsInstance(pointer, model.Pointer)
        self.assertEqual(pointer.value, b"@pointer1@")
        self.assertEqual(pointer.ref, "1")

    def test_041_Pointer_sub(self):
        """Test Pointer class."""

        class Parser(object):
            def __init__(self):
                self.xref0 = {b"@pointer0@": (0, "TAG0"),
                              b"@pointer1@": (1, "TAG1")}

            def read_record(self, offset):
                return str(offset)

        dialect = model.DIALECT_MYHERITAGE
        parser = Parser()
        husb = model.make_record(1, None, "HUSB", b"@pointer0@", [], 0, dialect, parser).freeze()
        wife = model.make_record(1, None, "WIFE", b"@pointer1@", [], 0, dialect, parser).freeze()
        fam = model.make_record(0, None, "FAM", "", [husb, wife], 0, dialect, parser).freeze()

        rec = fam.sub_tag("HUSB", follow=True)
        self.assertEqual(rec, "0")
        rec = fam.sub_tag("HUSB", follow=False)
        self.assertTrue(rec is not None)
        self.assertIsInstance(rec, model.Pointer)

        recs = fam.sub_tags("HUSB", "WIFE", follow=True)
        self.assertEqual(recs, ["0", "1"])
        recs = fam.sub_tags("HUSB", "WIFE", follow=False)
        self.assertEqual(len(recs), 2)
        self.assertIsInstance(recs[0], model.Pointer)
        self.assertIsInstance(recs[1], model.Pointer)

    def test_900_make_record(self):
        """Test make_record method()"""

        rec = model.make_record(0, "@xref@", "TAG", "value", [], 1000,
                                model.DIALECT_DEFAULT)
        rec.freeze()
        self.assertTrue(type(rec) is model.Record)
        self.assertEqual(rec.level, 0)
        self.assertEqual(rec.xref_id, "@xref@")
        self.assertEqual(rec.tag, "TAG")
        self.assertEqual(rec.value, "value")
        self.assertEqual(rec.sub_records, [])
        self.assertEqual(rec.offset, 1000)
        self.assertEqual(rec.dialect, model.DIALECT_DEFAULT)

        rec = model.make_record(1, None, "NAME", "Joe", [], 1000,
                                model.DIALECT_ALTREE)
        rec.freeze()
        self.assertTrue(type(rec) is model.NameRec)
        self.assertEqual(rec.level, 1)
        self.assertTrue(rec.xref_id is None)
        self.assertEqual(rec.tag, "NAME")
        self.assertEqual(rec.value, ("Joe", "", ""))
        self.assertEqual(rec.sub_records, [])
        self.assertEqual(rec.offset, 1000)
        self.assertEqual(rec.dialect, model.DIALECT_ALTREE)

        rec = model.make_record(0, "@I1@", "INDI", None, [], 1000,
                                model.DIALECT_MYHERITAGE)
        rec.freeze()
        self.assertTrue(type(rec) is model.Individual)
        self.assertEqual(rec.level, 0)
        self.assertEqual(rec.xref_id, "@I1@")
        self.assertEqual(rec.tag, "INDI")
        self.assertTrue(rec.value is None)
        self.assertEqual(rec.sub_records, [])
        self.assertEqual(rec.offset, 1000)
        self.assertEqual(rec.dialect, model.DIALECT_MYHERITAGE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ged4py.parser` module."""

from contextlib import contextmanager
import io
import tempfile
import os
import unittest


from ged4py import model, parser


@contextmanager
def _temp_file(data):
    """Create file with unique name and store some data in it.

    Returns file name.
    """
    fd, fname = tempfile.mkstemp()
    os.write(fd, data)
    os.close(fd)
    yield fname
    os.unlink(fname)


class TestParser(unittest.TestCase):
    """Tests for `ged4py.parser` module."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_002_guess_codec(self):
        """Test guess_codec()."""

        file = io.BytesIO(b"0 HEAD\n1 CHAR ASCII\n0 TRLR")
        self.assertEqual(parser.guess_codec(file), ("ascii", 0))

        file = io.BytesIO(b"0 HEAD\n1 CHAR ANSEL\n0 TRLR")
        self.assertEqual(parser.guess_codec(file), ("gedcom", 0))

        file = io.BytesIO(b"0 HEAD\n1 CHAR UTF-8\n0 TRLR")
        self.assertEqual(parser.guess_codec(file), ("utf-8", 0))

        file = io.BytesIO(b"\xef\xbb\xbf0 HEAD\n1 CHAR UTF-8\n0 TRLR")
        self.assertEqual(parser.guess_codec(file), ("utf-8", 3))

        # CR-LF
        file = io.BytesIO(b"\xef\xbb\xbf0 HEAD\r\n1 CHAR UTF-8\r\n0 TRLR")
        self.assertEqual(parser.guess_codec(file), ("utf-8", 3))

#         UTF-16 is broken
#         # utf-16-le
#         file = io.BytesIO(b"\xff\xfe0\0 \0H\0E\0A\0D\n\0\x31\0 \0C\0H\0A\0R\0 \0U\0T\0F\0-\01\06\0")
#         self.assertEqual(parser.guess_codec(file), "utf-16")
#
#         # utf-16-be
#         file = io.BytesIO(b"\xfe\xff\0\x30\0 \0H\0E\0A\0D")
#         self.assertEqual(parser.guess_codec(file), "utf-16")

    def test_003_codec_exceptions(self):
        """Test codecs-related exceptions."""

        # unknown codec name
        file = io.BytesIO(b"0 HEAD\n1 CHAR NOTCODEC\n0 TRLR")
        self.assertRaises(parser.CodecError, parser.guess_codec, file)

        # BOM disagrees with CHAR
        file = io.BytesIO(b"\xef\xbb\xbf0 HEAD\n1 CHAR ANSEL\n0 TRLR")
        self.assertRaises(parser.CodecError, parser.guess_codec, file)

    def test_010_open(self):
        """Test gedcom_open() method."""

        data = b"0 HEAD\n1 CHAR ANSEL\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader._encoding, "gedcom")
                self.assertEqual(reader._bom_size, 0)

        data = b"0 HEAD\n1 CHAR ASCII\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader._encoding, "ascii")
                self.assertEqual(reader._bom_size, 0)

        data = b"0 HEAD\n1 CHAR UTF-8\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader._encoding, "utf-8")
                self.assertEqual(reader._bom_size, 0)

        data = b"\xef\xbb\xbf0 HEAD\n1 CHAR UTF-8\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader._encoding, "utf-8")
                self.assertEqual(reader._bom_size, 3)

        # CR-LF
        data = b"\xef\xbb\xbf0 HEAD\r\n1 CHAR UTF-8\r\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader._encoding, "utf-8")
                self.assertEqual(reader._bom_size, 3)

        data = b"0 HEAD\n1 CHAR ASCII\n0 TRLR"
        with io.BytesIO(data) as file:
            with parser.GedcomReader(file) as reader:
                self.assertEqual(reader._encoding, "ascii")
                self.assertEqual(reader._bom_size, 0)

        data = b"\xef\xbb\xbf0 HEAD\n1 CHAR UTF-8\n0 TRLR"
        with io.BytesIO(data) as file:
            with parser.GedcomReader(file) as reader:
                self.assertEqual(reader._encoding, "utf-8")
                self.assertEqual(reader._bom_size, 3)

        data = b"0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader._encoding, "gedcom")
                self.assertEqual(reader._bom_size, 0)

        data = b"\xef\xbb\xbf0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader._encoding, "utf-8")
                self.assertEqual(reader._bom_size, 3)

    def test_011_open_errors(self):
        """Test gedcom_open() method."""

        # no HEAD
        data = b"\xef\xbb\xbf0 HDR\n1 CHAR ANSEL\n0 TRLR"
        with _temp_file(data) as fname:
            self.assertRaises(parser.CodecError, parser.GedcomReader, fname,
                              require_char=True)

        # no CHAR
        data = b"\xef\xbb\xbf0 HEAD\n1 NOCHAR ANSEL\n0 TRLR"
        with _temp_file(data) as fname:
            self.assertRaises(parser.CodecError, parser.GedcomReader, fname,
                              require_char=True)

        # unknown encoding
        data = b"\xef\xbb\xbf0 HEAD\n1 CHAR not-an-encoding\n0 TRLR"
        with _temp_file(data) as fname:
            self.assertRaises(parser.CodecError, parser.GedcomReader, fname)

        # expect UTF-8
        data = b"\xef\xbb\xbf0 HEAD\n1 CHAR ANSEL\n0 TRLR"
        with _temp_file(data) as fname:
            self.assertRaises(parser.CodecError, parser.GedcomReader, fname)

    def test_015_init_index(self):
        """Test _init_index() method."""

        data = b"0 HEAD\n0 @i1@ INDI\n0 @i2@ INDI\n0 @i3@ INDI\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertTrue(reader._index0 is None)
                self.assertTrue(reader._xref0 is None)
                reader._init_index()
                self.assertEqual(reader._index0, [(0, "HEAD"),
                                                  (7, "INDI"),
                                                  (19, "INDI"),
                                                  (31, "INDI"),
                                                  (43, "TRLR")])
                self.assertEqual(reader._xref0, {"@i1@": (7, "INDI"),
                                                 "@i2@": (19, "INDI"),
                                                 "@i3@": (31, "INDI")})

    def test_017_dialect(self):
        """Test dialect property."""

        data = b"0 HEAD\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader.dialect, model.DIALECT_DEFAULT)

        data = b"0 HEAD\n1 SOUR MYHERITAGE\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader.dialect, model.DIALECT_MYHERITAGE)

        data = b"0 HEAD\n1 SOUR ALTREE\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader.dialect, model.DIALECT_ALTREE)

        data = b"0 HEAD\n1 SOUR ANCESTRIS\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader.dialect, model.DIALECT_ANCESTRIS)

        data = b"0 HEAD\n1 SOUR XXX\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                self.assertEqual(reader.dialect, model.DIALECT_DEFAULT)

    def test_020_gedcom_lines(self):
        """Test gedcom_lines method"""

        # simple content
        data = b"0 HEAD\n1 CHAR ASCII\n1 SOUR PIF PAF\n0 @i1@ INDI\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                lines = list(reader.gedcom_lines(0))
                expect = [parser.gedcom_line(level=0, xref_id=None, tag="HEAD", value=None, offset=0),
                          parser.gedcom_line(level=1, xref_id=None, tag="CHAR", value=b"ASCII", offset=7),
                          parser.gedcom_line(level=1, xref_id=None, tag="SOUR", value=b"PIF PAF", offset=20),
                          parser.gedcom_line(level=0, xref_id="@i1@", tag="INDI", value=None, offset=35),
                          parser.gedcom_line(level=0, xref_id=None, tag="TRLR", value=None, offset=47)]
                self.assertEqual(lines, expect)

        # Unicode characters
        data = b"0 HEAD\n1 CHAR UTF-8\n0 OK \xc2\xb5"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                lines = list(reader.gedcom_lines(0))
                expect = [parser.gedcom_line(level=0, xref_id=None, tag="HEAD", value=None, offset=0),
                          parser.gedcom_line(level=1, xref_id=None, tag="CHAR", value=b"UTF-8", offset=7),
                          parser.gedcom_line(level=0, xref_id=None, tag="OK", value=b"\xc2\xb5", offset=20)]
                self.assertEqual(lines, expect)

        # Unicode and BOM
        data = b"\xef\xbb\xbf0 HEAD\r\n1 CHAR UTF-8\r\n0 OK \xc2\xb5"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                lines = list(reader.gedcom_lines(3))
                expect = [parser.gedcom_line(level=0, xref_id=None, tag="HEAD", value=None, offset=3),
                          parser.gedcom_line(level=1, xref_id=None, tag="CHAR", value=b"UTF-8", offset=11),
                          parser.gedcom_line(level=0, xref_id=None, tag="OK", value=b"\xc2\xb5", offset=25)]
                self.assertEqual(lines, expect)

        data = b"\xef\xbb\xbf0 HEAD\r\n1 CHAR UTF-8\r\n0 OK \xc2\xb5"
        with io.BytesIO(data) as file:
            with parser.GedcomReader(file) as reader:
                lines = list(reader.gedcom_lines(3))
                expect = [parser.gedcom_line(level=0, xref_id=None, tag="HEAD", value=None, offset=3),
                          parser.gedcom_line(level=1, xref_id=None, tag="CHAR", value=b"UTF-8", offset=11),
                          parser.gedcom_line(level=0, xref_id=None, tag="OK", value=b"\xc2\xb5", offset=25)]
                self.assertEqual(lines, expect)

    def test_021_gedcom_lines_errors(self):
        """Test for exceptions raised by gedcom_lines method"""

        # tag name is only letters and digits
        data = b"0 HEAD\n1 CHAR ASCII\n1 SO@UR PIF PAF"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                itr = reader.gedcom_lines(0)
                self.assertRaises(parser.ParserError, list, itr)

        # xref must start with letter or digit
        data = b"0 HEAD\n1 CHAR ASCII\n1 @!ref@ SOUR PIF PAF"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                itr = reader.gedcom_lines(0)
                self.assertRaises(parser.ParserError, list, itr)

        # level must be a number
        data = b"0 HEAD\n1 CHAR ASCII\nX SOUR PIF PAF"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                itr = reader.gedcom_lines(0)
                self.assertRaises(parser.ParserError, list, itr)

        # consistency check - nested levels
        datas = [b"0 HEAD\n1 CHAR ASCII\n1 DATA\n3 MORE DATA",
                 b"0 HEAD\n1 CHAR ASCII\n1 DATA adata\n1 CONC aadata",
                 b"0 HEAD\n1 CHAR ASCII\n1 DATA adata\n3 CONC aadata",
                 b"0 HEAD\n1 CHAR ASCII\n1 DATA adata\n1 CONT aadata",
                 b"0 HEAD\n1 CHAR ASCII\n1 DATA adata\n3 CONT aadata",
                 b"0 HEAD\n1 CHAR ASCII\n1 DATA adata\n2 CONT aadata\n3 CONT aadata",
                 b"0 HEAD\n1 CHAR ASCII\n1 DATA adata\n2 CONT aadata\n1 CONT aadata",
                 ]
        for data in datas:
            print(data)
            with _temp_file(data) as fname:
                with parser.GedcomReader(fname) as reader:
                    itr = reader.gedcom_lines(0)
                    self.assertRaises(parser.IntegrityError, list, itr)

    def test_030_read_record(self):
        """Test read_record method"""

        data = b"0 HEAD\n1 CHAR ASCII\n0 INDI A\n0 INDI B"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                rec = reader.read_record(20)
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "INDI")
                self.assertEqual(rec.value, "A")
                self.assertEqual(rec.sub_records, [])

                rec = reader.read_record(29)
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "INDI")
                self.assertEqual(rec.value, "B")
                self.assertEqual(rec.sub_records, [])

        data = b"0 HEAD\n1 CHAR ASCII\n0 INDI A\n1 SUBA A\n1 SUBB B\n2 SUBC C\n1 SUBD D\n0 STOP"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                rec = reader.read_record(20)
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "INDI")
                self.assertEqual(rec.value, "A")
                self.assertEqual(len(rec.sub_records), 3)

                suba = rec.sub_records[0]
                self.assertEqual(suba.level, 1)
                self.assertEqual(suba.tag, "SUBA")
                self.assertEqual(suba.value, "A")
                self.assertEqual(len(suba.sub_records), 0)

                subb = rec.sub_records[1]
                self.assertEqual(subb.level, 1)
                self.assertEqual(subb.tag, "SUBB")
                self.assertEqual(subb.value, "B")
                self.assertEqual(len(subb.sub_records), 1)

                subc = subb.sub_records[0]
                self.assertEqual(subc.level, 2)
                self.assertEqual(subc.tag, "SUBC")
                self.assertEqual(subc.value, "C")
                self.assertEqual(len(subc.sub_records), 0)

                subd = rec.sub_records[2]
                self.assertEqual(subd.level, 1)
                self.assertEqual(subd.tag, "SUBD")
                self.assertEqual(subd.value, "D")
                self.assertEqual(len(subd.sub_records), 0)

        data = b"0 HEAD\n1 CHAR ASCII\n0 INDI A\n1 NOTE A\n2 CONC B\n2 CONT C\n2 CONC D"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                rec = reader.read_record(20)
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "INDI")
                self.assertEqual(rec.value, "A")
                self.assertEqual(len(rec.sub_records), 1)

                note = rec.sub_records[0]
                self.assertEqual(note.level, 1)
                self.assertEqual(note.tag, "NOTE")
                self.assertEqual(note.value, "AB\nCD")
                self.assertEqual(len(note.sub_records), 0)

        # Space-aware concatenation
        data = b"0 HEAD\n1 CHAR ASCII\n0 INDI A\n1 NOTE\n2 CONC B\n2 CONT C\n2 CONC  D"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                note = reader.read_record(29)
                self.assertEqual(note.level, 1)
                self.assertEqual(note.tag, "NOTE")
                self.assertEqual(note.value, "B\nC D")
                self.assertEqual(len(note.sub_records), 0)

        # BLOB
        data = b"0 HEAD\n1 CHAR ASCII\n0 INDI A\n1 BLOB\n2 CONT A\n2 CONT B"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                note = reader.read_record(29)
                self.assertEqual(note.level, 1)
                self.assertEqual(note.tag, "BLOB")
                self.assertIsNone(note.value)
                self.assertEqual(len(note.sub_records), 0)

    def test_031_read_record_conc(self):
        # encoded string
        data = b"0 HEAD\n1 CHAR UTF8\n"\
            b"0 TAG \xd0\x98\xd0\xb2\xd0\xb0\xd0\xbd \xd0\x98\xd0\xb2\xd0\xb0"\
            b"\xd0\xbd\xd0\xbe\xd0\xb2\xd0\xb8\xd1\x87"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                note = reader.read_record(19)
                self.assertEqual(note.level, 0)
                self.assertEqual(note.tag, "TAG")
                self.assertEqual(note.value, u"Иван Иванович")

        # encoded string split
        data = b"0 HEAD\n1 CHAR UTF8\n"\
            b"0 TAG \xd0\x98\xd0\xb2\xd0\xb0\xd0\xbd \xd0\x98\n"\
            b"1 CONC \xd0\xb2\xd0\xb0\xd0\xbd\xd0\xbe\xd0\xb2\xd0\xb8\xd1\x87"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                note = reader.read_record(19)
                self.assertEqual(note.level, 0)
                self.assertEqual(note.tag, "TAG")
                self.assertEqual(note.value, u"Иван Иванович")

        # encoded string split in the middle of multi-byte character
        data = b"0 HEAD\n1 CHAR UTF8\n"\
            b"0 TAG \xd0\x98\xd0\xb2\xd0\xb0\xd0\xbd \xd0\n"\
            b"1 CONC \x98\xd0\xb2\xd0\xb0\xd0\xbd\xd0\xbe\xd0\xb2\xd0\xb8\xd1\x87"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                note = reader.read_record(19)
                self.assertEqual(note.level, 0)
                self.assertEqual(note.tag, "TAG")
                self.assertEqual(note.value, u"Иван Иванович")

        # encoded string split in between combining characters (UTF-8)
        data = b"0 HEAD\n1 CHAR UTF8\n"\
            b"0 TAG Pa\n"\
            b"1 CONC \xcc\x8al"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                note = reader.read_record(19)
                self.assertEqual(note.level, 0)
                self.assertEqual(note.tag, "TAG")
                self.assertEqual(note.value, u"Pål")

        # encoded string split in between combining characters (ANSEL)
        data = b"0 HEAD\n1 CHAR ANSEL\n"\
            b"0 TAG P\xea\n"\
            b"1 CONC al"

        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                note = reader.read_record(20)
                self.assertEqual(note.level, 0)
                self.assertEqual(note.tag, "TAG")
                self.assertEqual(note.value, u"Pål")

    def test_035_read_record_errors(self):
        """Test read_record method"""

        data = b"0 HEAD\n1 CHAR ASCII\n0 INDI A\n0 INDI B"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:

                # try to read beyond EOF should return None
                rec = reader.read_record(1000)
                self.assertTrue(rec is None)

                # Random location
                self.assertRaises(parser.ParserError, reader.read_record, 26)

    def test_040_records0(self):
        """Test records0 method"""

        data = b"0 HEAD\n1 CHAR ASCII\n0 INDI A\n1 SUBA A\n1 SUBB B\n2 SUBC C\n1 SUBD D\n0 STOP"
        with io.BytesIO(data) as file:
            with parser.GedcomReader(file) as reader:

                recs = list(reader.records0())
                self.assertEqual(len(recs), 3)

                rec = recs[0]
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "HEAD")
                self.assertEqual(rec.value, None)
                self.assertEqual(len(rec.sub_records), 1)
                self.assertEqual(rec.dialect, model.DIALECT_DEFAULT)

                rec = recs[1]
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "INDI")
                self.assertEqual(rec.value, "A")
                self.assertEqual(len(rec.sub_records), 3)
                self.assertEqual(rec.dialect, model.DIALECT_DEFAULT)

                rec = recs[2]
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "STOP")
                self.assertEqual(rec.value, None)
                self.assertEqual(len(rec.sub_records), 0)
                self.assertEqual(rec.dialect, model.DIALECT_DEFAULT)

    def test_041_header(self):
        """Test header property."""

        data = b"0 HEAD\n1 SOUR ALTREE\n0 TRLR"
        with _temp_file(data) as fname:
            with parser.GedcomReader(fname) as reader:
                rec = reader.header
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "HEAD")
                self.assertEqual(rec.value, None)
                self.assertEqual(len(rec.sub_records), 1)
                self.assertEqual(rec.dialect, model.DIALECT_DEFAULT)

    def test_042_rec_dialect(self):
        """Test records0 method"""

        data = b"0 HEAD\n1 CHAR ASCII\n1 SOUR ALTREE\n"\
            b"0 INDI A\n1 SUBA A\n1 SUBB B\n2 SUBC C\n1 SUBD D\n0 STOP"
        with io.BytesIO(data) as file:
            with parser.GedcomReader(file) as reader:

                recs = list(reader.records0())
                self.assertEqual(len(recs), 3)

                rec = recs[0]
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "HEAD")
                self.assertEqual(rec.value, None)
                self.assertEqual(len(rec.sub_records), 2)
                self.assertEqual(rec.dialect, model.DIALECT_DEFAULT)

                rec = recs[1]
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "INDI")
                self.assertEqual(rec.value, "A")
                self.assertEqual(len(rec.sub_records), 3)
                self.assertEqual(rec.dialect, model.DIALECT_ALTREE)

                rec = recs[2]
                self.assertEqual(rec.level, 0)
                self.assertEqual(rec.tag, "STOP")
                self.assertEqual(rec.value, None)
                self.assertEqual(len(rec.sub_records), 0)
                self.assertEqual(rec.dialect, model.DIALECT_ALTREE)



# -*- coding: utf-8 -*-

"""Module containing Python in-memory model for GEDCOM data.
"""

from __future__ import print_function, absolute_import, division

__all__ = ['make_record', 'Record', 'Pointer', 'NameRec', 'Name',
           'Date', 'Individual']

from .detail.name import (split_name, parse_name_altree, parse_name_ancestris,
                          parse_name_myher)
from .detail.date import DateValue

# Even though the structure of GEDCOM file is more or less fixed,
# interpretation of some data may vary depending on which application
# produced GEDCOM file. Constants define different known dialect which
# are handled by classes below.
DIALECT_DEFAULT = "DEF"
DIALECT_MYHERITAGE = "MYHER"  # myheritage.com
DIALECT_ALTREE = "AGELONG"  # Agelong Tree (genery.com)
DIALECT_ANCESTRIS = "ANCESTRIS"  # Ancestris (ancestris.org)

# Names/Individuals can be ordered differently, e.g. by surname first,
# by given name first, or by maiden name first. This few constants define
# different ordering options.
ORDER_SURNAME_GIVEN = "last+first"
ORDER_GIVEN_SURNAME = "first+last"
ORDER_MAIDEN_GIVEN = "maiden+first"  # uses last name if no maiden name
ORDER_GIVEN_MAIDEN = "first+maiden"  # uses last name if no maiden name
ORDER_LIST = [ORDER_SURNAME_GIVEN, ORDER_GIVEN_SURNAME,
              ORDER_MAIDEN_GIVEN, ORDER_GIVEN_MAIDEN]


class Record(object):
    """Class representing a parsed GEDCOM record in a generic format.

    Client code usually does not need to create instances of this class
    directly, :py:meth:`make_record` should be used instead. If you create
    an instance of this class (or its subclass) then you are responsible for
    filling its attributes.

    :ivar int level: Record level number
    :ivar str xref_id: Record reference ID, possibly empty.
    :ivar str tag: Tag name
    :ivar object value: Record value, possibly None, for many record types
        value is a string or None, some subclasses can define different type
        of record value.
    :ivar list sub_records: List of subordinate records, possibly empty
    :ivar int offset: Record location in a file
    :ivar dialect: GEDCOM source dialect, one of the DIALECT_* values
    """
    def __init__(self):
        self.level = None
        self.xref_id = None
        self.tag = None
        self.value = None
        self.sub_records = None
        self.offset = None
        self.dialect = None

    def freeze(self):
        """Method called by parser when updates to this record finish.

        :return: self
        """
        return self

    def sub_tag(self, path, follow=True):
        """Returns direct sub-record with given tag name or None.

        Path can be a simple tag name, in which case the first direct
        sub-record of this record with the matching tag is returned. Path
        can also consist of several tags separated by slashes, in that case
        sub-records are searched recursively.

        If `follow` is True then pointer records are resolved and pointed
        record is used instead of pointer record, this also works for all
        intermediate records in a path.

        :param str path: tag names separated by slashes.
        :param boolean follow: If True then resolve pointers.
        :return: `Record` instance or `None` if sub-record with a given
            tag does not exist.
        """
        tags = path.split('/')
        rec = self
        for tag in tags:
            recs = [x for x in (rec.sub_records or []) if x.tag == tag]
            if not recs:
                return None
            rec = recs[0]
            if follow and isinstance(rec, Pointer):
                rec = rec.ref
        return rec

    def sub_tag_value(self, path, follow=True):
        """Returns value of a direct sub-record or None.

        Works as :py:meth:`sub_tag` but returns value of a sub-record
        instead of sub-record itself.

        :param str path: tag names separated by slashes.
        :param boolean follow: If True then resolve pointers.
        :return: String or `None` if sub-record with a given
            tag does not exist.
        """
        rec = self.sub_tag(path, follow)
        if rec:
            return rec.value
        return None

    def sub_tags(self, *tags, **kw):
        """Returns list of direct sub-records matching any tag name.

        Unlike :py:meth:`sub_tag` method this method does not support
        hierarchical paths and does not resolve pointers.

        :param str tags: Names of the sub-record tag
        :param kw: Keyword arguments, only recognized keyword is `follow`
            with the same meaning as in :py:meth:`sub_tag`.
        :return: List of `Records`, possibly empty.
        """
        records = [x for x in self.sub_records if x.tag in tags]
        if kw.get('follow', True):
            records = [rec.ref if isinstance(rec, Pointer) else rec
                       for rec in records]
        return records

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        value = self.value
        if isinstance(value, (type(""), type(u""))) and len(value) > 32:
            value = value[:32] + "..."
        n_sub = len(self.sub_records)
        if self.xref_id:
            fmt = "{0}(level={1.level}, xref_id={1.xref_id}, tag={1.tag}, " \
                "value={2!r}, offset={1.offset}, #subrec={3})"
        else:
            fmt = "{0}(level={1.level}, tag={1.tag}, " \
                "value={2!r}, offset={1.offset}, #subrec={3})"
        return fmt.format(self.__class__.__name__, self, value, n_sub)


class Pointer(Record):
    """Class representing a reference to a record in a GEDCOM file.

    This class wraps a GEDCOM pointer and adds few useful methods to locate
    and retrieve a pointed object. Instance of this class will be used in
    place of the GEDCOM pointers in the objects created by parser.

    :param parser: Instance of `GedcomReader` class.

    :ivar str value: Value of the GEDCOM pointer (e.g. "@I1234@")
    :ivar Record ref: dereferenced GEDCOM record
    """

    def __init__(self, parser):
        self.parser = parser
        self._value = []  # use non-None to signify non-initialized

    @property
    def ref(self):
        if self._value == []:
            offset, _ = self.parser.xref0.get(self.value, (None, None))
            if offset is None:
                self._value = None
            else:
                self._value = self.parser.read_record(offset)
        return self._value


class NameRec(Record):
    """Representation of the NAME record.

    This class adds few convenience methods for name manipulation. It also
    redefines the type of the `value` attribute, it's type is tuple.
    Value tuple can contain 3 or 4 elements, if there are 4 elements then
    last element is a maiden name. Second element of a tuple is surname,
    first and third elements are pieces of the given name (this is determined
    entirely by how name is represented in GEDCOM file). Any of the elements
    can be empty string. If NAME record value is empty in GEDCOM file then
    all three fields of the tuple will be empty strings. Few examples::

        ("John", "Smith", "")
        ("Mary Joan", "Smith", "", "Ivanova")    # maiden name
        ("", "Ivanov", "Ivan Ivanovich")
        ("John", "Smith", "Jr.")
        ("", "", "")                             # empty NAME record

    Client code usually does not need to create instances of this class
    directly, :py:meth:`make_record` should be used instead.
    """

    def __init__(self):
        Record.__init__(self)

    def freeze(self):
        """Method called by parser when updates to this record finish.

        :return: self
        """
        # None is the same as empty string
        if self.value is None:
            self.value = ""
        if self.dialect in [DIALECT_ALTREE]:
            name_tuple = parse_name_altree(self)
        elif self.dialect in [DIALECT_MYHERITAGE]:
            name_tuple = parse_name_myher(self)
        elif self.dialect in [DIALECT_ANCESTRIS]:
            name_tuple = parse_name_ancestris(self)
        else:
            name_tuple = split_name(self.value)
        self.value = name_tuple
        return self

    @property
    def type(self):
        """Name type as defined in TYPE record. ``None`` if TYPE record is
        missing, otherwise string, e.g. "aka", "birth", "immigrant",
        "maiden", "married" (or anything else).
        """
        # +1 TYPE <NAME_TYPE> {0:1}
        rec = self.sub_tag("TYPE")
        return rec.value if rec else None

    def __str__(self):
        return Record.__str__(self)


class Name(object):
    """Class representing summary of person names.

    Person in GEDCOM can have multiple NAME records, e.g. "aka" name,
    "maiden" name, etc. This class provides simple interface for accessing
    info from all those records.

    :param list names: List of NAME records (:py:class:`NameRec` instances).
    :param dialect: One of DIALECT_* constants.
    """

    def __init__(self, names, dialect):
        self._names = names
        self._dialect = dialect
        self._primary = None  # "primary" name record

        if len(names) == 0:
            # make fake name record to simplify logic below
            self._primary = make_record(0, '', "NAME", "",
                                        [], 0, DIALECT_DEFAULT).freeze()
        elif len(names) == 1:
            self._primary = names[0]
        else:
            for name in names:
                if not name.type:
                    self._primary = name
                    break
            else:
                self._primary = names[0]

    @property
    def surname(self):
        return self._primary.value[1]

    @property
    def given(self):
        """Given name could include both first and middle name"""
        if self._primary.value[0] and self._primary.value[2]:
            return self._primary.value[0] + ' ' + self._primary.value[2]
        return self._primary.value[0] or self._primary.value[2]

    @property
    def first(self):
        """First name is the first part of a given name (drops middle name)"""
        given = self.given
        if given:
            return given.split()[0]
        return given

    @property
    def maiden(self):
        """Maiden last name, can be None"""
        if self._dialect == DIALECT_DEFAULT:
            # for default/unknown dialect try "maiden" name record first
            for name in self._names:
                if name.type == "maiden":
                    return name.value[1]
        # rely on NameRec extracting it from other source
        if self._primary and len(self._primary.value) > 3:
            return self._primary.value[3]
        return None

    def order(self, order):
        """Returns name order key.

        Returns tuple with two strings that can be compared to other such
        tuple obtained from different name. Note that if you want
        locale-dependent ordering then you need to compare strings using
        locale-aware method (e.g. ``locale.strxfrm``).

        :param order: One of the ORDER_* constants.
        :returns: tuple of two strings
        """
        given = self.given
        surname = self.surname
        if order in (ORDER_MAIDEN_GIVEN, ORDER_GIVEN_MAIDEN):
            surname = self.maiden or self.surname

        # We are collating empty names to come after non-empty,
        # so instead of empty we return "2" and add "1" as prefix to others
        given = ("1" + given) if given else "2"
        surname = ("1" + surname) if surname else "2"

        if order in (ORDER_SURNAME_GIVEN, ORDER_MAIDEN_GIVEN):
            return (surname, given)
        elif order in (ORDER_GIVEN_SURNAME, ORDER_GIVEN_MAIDEN):
            return (given, surname)
        else:
            raise ValueError("unexpected order: {}".format(order))

    def format(self):
        """Format name for output.

        :return: Formatted name representation.
        """
        name = self._primary.value[0]
        if self.surname:
            if name:
                name += ' '
            name += self.surname
        if self._primary.value[2]:
            if name:
                name += ' '
            name += self._primary.value[2]
        return name

    def __str__(self):
        fmt = "{0}({1!r})"
        return fmt.format(self.__class__.__name__, self.format())


class Date(Record):
    """Representation of the DATE record.

    After `freeze()` method is called by parser the `value` attribute contains
    instance of :py:class:`ged4py.detail.date.DateValue` class.
    """

    def freeze(self):
        """Method called by parser when updates to this record finish.

        :return: self
        """
        self.value = DateValue.parse(self.value)
        return self


class Individual(Record):
    """Representation of the INDI record.

    INDI record represents a single person in GEDCOM. This class defines
    few methods that may be useful for manipulating person records, such
    as ordering, navigation, etc.

    Client code usually does not need to create instances of this class
    directly, :py:meth:`make_record` should be used instead.
    """

    def __init__(self):
        Record.__init__(self)
        self._mother = []  # Non-None as uninitialized
        self._father = []  # Non-None as uninitialized

    @property
    def name(self):
        """:py:class:`Name` instance.
        """
        # +1 <<PERSONAL_NAME_STRUCTURE>> {0:M}
        return Name(self.sub_tags("NAME"), self.dialect)

    @property
    def sex(self):
        """Person sex, "M", "F", or "U"."""
        # +1 SEX <SEX_VALUE>
        sex_rec = self.sub_tag("SEX")
        if sex_rec:
            return sex_rec.value
        return "U"

    @property
    def mother(self):
        """Parent of this individual"""
        if self._mother == []:
            self._mother = self.sub_tag("FAMC/WIFE")
        return self._mother

    @property
    def father(self):
        """Parent of this individual"""
        if self._father == []:
            self._father = self.sub_tag("FAMC/HUSB")
        return self._father


# maps tag names to record class
_tag_class = dict(INDI=Individual,
                  NAME=NameRec,
                  DATE=Date)


def make_record(level, xref_id, tag, value, sub_records, offset, dialect,
                parser=None):
    """Create Record instance based on parameters.

    :param int level: Record level number.
    :param str xref_id: Record reference ID, possibly empty.
    :param str tag: Tag name.
    :param value: Record value, possibly empty. Value can be None, bytes, or
        string object, if it is bytes then it should be decoded into strings
        before calling freeze(), this is normally done by the parser which
        knows about encodings.
    :param list sub_records: Initial list of subordinate records,
        possibly empty. List can be updated later.
    :param int offset: Record location in a file.
    :param dialect: One of DIALECT_* constants.
    :param parser: Instance of `GedcomReader` class, only needed for
        records whose walue is a pointer.
    :return: Instance of :py:class:`Record` (or one of its subclasses).
    """
    # value can be bytes or string so we check for both, 64 is code for '@'
    if value and len(value) > 2 and \
        ((value[0] == '@' and value[-1] == '@') or
         (value[0] == 64 and value[-1] == 64)):
        # this looks like a <pointer>, make a Pointer record
        klass = Pointer
        rec = klass(parser)
    else:
        klass = _tag_class.get(tag, Record)
        rec = klass()

    rec.level = level
    rec.xref_id = xref_id
    rec.tag = tag
    rec.value = value
    rec.sub_records = sub_records
    rec.offset = offset
    rec.dialect = dialect
    return rec



# -*- coding: utf-8 -*-

"""Top-level package for GEDCOM parser for Python."""

__author__ = """Andy Salnikov"""
__email__ = 'ged4py@py-dev.com'
__version__ = '0.1.12'

import ansel as _ansel
_ansel.register()

from .parser import GedcomReader  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'ansel'
]

setup_requirements = [
]

test_requirements = [
]

setup(
    name='ged4py',
    version='0.1.12',
    description="GEDCOM tools for Python",
    long_description=readme + '\n\n' + history,

    author="Andy Salnikov",
    author_email='ged4py@py-dev.com',
    url='https://github.com/andy-z/ged4py',
    license="MIT License",

    packages=find_packages(include=['ged4py', 'ged4py.*']),
    entry_points={
        'console_scripts': [
        ]
    },
    include_package_data=False,

    install_requires=requirements,

    zip_safe=True,
    keywords='ged4py',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Sociology :: Genealogy',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ged4py.detail.date` module."""

import unittest

from ged4py.detail.date import CalendarDate, DateValue


class TestDetailName(unittest.TestCase):
    """Tests for `ged4py.detail.io` module."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_001_cal_date(self):
        """Test detail.date.CalendarDate class."""

        date = CalendarDate("2017", "OCT", 9, "GREGORIAN")
        self.assertEqual(date.year, "2017")
        self.assertEqual(date.month, "OCT")
        self.assertEqual(date.day, 9)
        self.assertEqual(date.calendar, "GREGORIAN")

        date = CalendarDate("2017B.C.", "OCT", None, None)
        self.assertEqual(date.year, "2017B.C.")
        self.assertEqual(date.month, "OCT")
        self.assertTrue(date.day is None)
        self.assertEqual(date.calendar, "GREGORIAN")

        date = CalendarDate("5000", calendar="HEBREW")
        self.assertEqual(date.year, "5000")
        self.assertTrue(date.month is None)
        self.assertTrue(date.day is None)
        self.assertEqual(date.calendar, "HEBREW")

        date = CalendarDate()
        self.assertTrue(date.year is None)
        self.assertTrue(date.month is None)
        self.assertTrue(date.day is None)
        self.assertEqual(date.calendar, "GREGORIAN")

    def test_002_cal_date_as_tuple(self):
        """Test detail.date.CalendarDate class."""

        date = CalendarDate("2017", "OCT", 9, "GREGORIAN")
        self.assertEqual(date.as_tuple, (2017, 10, 9))

        date = CalendarDate("2017B.C.", "VENT", None, "FRENCH R")
        self.assertEqual(date.as_tuple, (2017, 6, 99))

        date = CalendarDate("2017B.C.", "TSH", 22, "HEBREW")
        self.assertEqual(date.as_tuple, (2017, 1, 22))

        date = CalendarDate("5000")
        self.assertEqual(date.as_tuple, (5000, 99, 99))

        date = CalendarDate()
        self.assertEqual(date.as_tuple, (9999, 99, 99))

    def test_003_cal_date_cmp(self):
        """Test detail.date.CalendarDate class."""

        self.assertTrue(CalendarDate("2016", "JAN", 1) < CalendarDate("2017", "JAN", 1))
        self.assertTrue(CalendarDate("2017", "JAN", 1) < CalendarDate("2017", "FEB", 1))
        self.assertTrue(CalendarDate("2017", "JAN", 1) < CalendarDate("2017", "JAN", 2))

        self.assertTrue(CalendarDate("2017", "JAN", 1) <= CalendarDate("2017", "JAN", 2))
        self.assertTrue(CalendarDate("2017", "JAN", 2) > CalendarDate("2017", "JAN", 1))
        self.assertTrue(CalendarDate("2017", "JAN", 2) >= CalendarDate("2017", "JAN", 1))
        self.assertTrue(CalendarDate("2017", "JAN", 1) == CalendarDate("2017", "JAN", 1))
        self.assertTrue(CalendarDate("2017", "JAN", 1) != CalendarDate("2017", "JAN", 2))

        # missing day compares as "past" the last day of month, but before next month
        self.assertTrue(CalendarDate("2017", "JAN") > CalendarDate("2017", "JAN", 31))
        self.assertTrue(CalendarDate("2017", "JAN") < CalendarDate("2017", "FEB", 1))
        # missing month compares as "past" the last day of year, but before next year
        self.assertTrue(CalendarDate("2017") > CalendarDate("2017", "DEC", 31))
        self.assertTrue(CalendarDate("2017") < CalendarDate("2018", "JAN", 1))

    def test_004_cal_date_fmt(self):
        """Test detail.date.CalendarDate class."""
        date = CalendarDate("2017", "OCT", 9, "GREGORIAN")
        self.assertEqual(date.fmt(), "2017 OCT 9")

        date = CalendarDate("2017B.C.", "OCT", None, None)
        self.assertEqual(date.fmt(), "2017B.C. OCT")

        date = CalendarDate("5000", calendar="HEBREW")
        self.assertEqual(date.fmt(), "5000")

    def test_010_date(self):
        """Test detail.date.DateValue class."""

        date = DateValue()
        self.assertEqual(date.template, "")
        self.assertEqual(date.kw, {})

        date = DateValue("$date", {})
        self.assertEqual(date.template, "$date")
        self.assertEqual(date.kw, {})

        date = DateValue("FROM $date1 TO $date2",
                         {"date1": CalendarDate("2017"),
                          "date2": CalendarDate("2020")})
        self.assertEqual(date.template, "FROM $date1 TO $date2")
        self.assertEqual(date.kw, {"date1": CalendarDate("2017"),
                                   "date2": CalendarDate("2020")})
        self.assertEqual(date.fmt(), "FROM 2017 TO 2020")

        # "phrase" keyword corresponds to string
        date = DateValue("FROM $date1 TO ($phrase)",
                         {"date1": CalendarDate("2017"),
                          "phrase": "some day"})
        self.assertEqual(date.fmt(), "FROM 2017 TO (some day)")

    def test_011_date_fmt(self):
        """Test detail.date.DateValue class."""

        date = DateValue("date", {})
        self.assertEqual(date.fmt(), "date")

        date = DateValue("FROM $date1 TO $date2",
                         {"date1": CalendarDate("2017"),
                          "date2": CalendarDate("2020")})
        self.assertEqual(date.fmt(), "FROM 2017 TO 2020")

        date = DateValue("BET $date1 AND $date2",
                         {"date1": CalendarDate("2017", "JAN", 1),
                          "date2": CalendarDate("2020", "FLOR", 20)})
        self.assertEqual(date.fmt(), "BET 2017 JAN 1 AND 2020 FLOR 20")

    def test_012_date_parse_period(self):
        """Test detail.date.DateValue class."""

        date = DateValue.parse("FROM 1967")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "FROM $date")
        self.assertEqual(date.kw, {"date": CalendarDate("1967")})
        self.assertEqual(date.fmt(), "FROM 1967")

        date = DateValue.parse("TO 1 JAN 2017")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "TO $date")
        self.assertEqual(date.kw, {"date": CalendarDate("2017", "JAN", 1)})
        self.assertEqual(date.fmt(), "TO 2017 JAN 1")

        date = DateValue.parse("FROM 1920 TO 2000")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "FROM $date1 TO $date2")
        self.assertEqual(date.kw, {"date1": CalendarDate("1920"),
                                   "date2": CalendarDate("2000")})
        self.assertEqual(date.fmt(), "FROM 1920 TO 2000")

        date = DateValue.parse("from mar 1920 to 1 apr 2000")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "FROM $date1 TO $date2")
        self.assertEqual(date.kw, {"date1": CalendarDate("1920", "MAR"),
                                   "date2": CalendarDate("2000", "APR", 1)})

    def test_013_date_parse_range(self):
        """Test detail.date.DateValue class."""

        date = DateValue.parse("BEF 1967A.D.")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "BEFORE $date")
        self.assertEqual(date.kw, {"date": CalendarDate("1967A.D.")})
        self.assertEqual(date.fmt(), "BEFORE 1967A.D.")

        date = DateValue.parse("AFT 1 JAN 2017")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "AFTER $date")
        self.assertEqual(date.kw, {"date": CalendarDate("2017", "JAN", 1)})
        self.assertEqual(date.fmt(), "AFTER 2017 JAN 1")

        date = DateValue.parse("BET @#DJULIAN@ 1600 AND 2000")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "BETWEEN $date1 AND $date2")
        self.assertEqual(date.kw, {"date1": CalendarDate("1600"),
                                   "date2": CalendarDate("2000")})
        self.assertEqual(date.fmt(), "BETWEEN 1600 AND 2000")

        date = DateValue.parse("bet mar 1920 and apr 2000")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "BETWEEN $date1 AND $date2")
        self.assertEqual(date.kw, {"date1": CalendarDate("1920", "MAR"),
                                   "date2": CalendarDate("2000", "APR")})
        self.assertEqual(date.fmt(), "BETWEEN 1920 MAR AND 2000 APR")

    def test_014_date_parse_approx(self):
        """Test detail.date.DateValue class."""

        dates = {"500B.C.": CalendarDate("500B.C."),
                 "@#DGREGORIAN@ JAN 2017": CalendarDate("2017", "JAN"),
                 "31 JAN 2017": CalendarDate("2017", "JAN", 31)}

        approx = {"ABT": "ABOUT", "CAL": "CALCULATED", "EST": "ESTIMATED"}

        for appr, tmpl in approx.items():
            for datestr, value in dates.items():

                date = DateValue.parse(appr + " " + datestr)
                self.assertTrue(date is not None)
                self.assertEqual(date.template, tmpl + " $date")
                self.assertEqual(date.kw, {"date": value})

    def test_015_date_parse_phrase(self):
        """Test detail.date.DateValue class."""

        date = DateValue.parse("(some phrase)")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "($phrase)")
        self.assertEqual(date.kw, {"phrase": "some phrase"})

        date = DateValue.parse("INT 1967A.D. (some phrase)")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "INTERPRETED $date ($phrase)")
        self.assertEqual(date.kw, {"date": CalendarDate("1967A.D."),
                                   "phrase": "some phrase"})
        self.assertEqual(date.fmt(), "INTERPRETED 1967A.D. (some phrase)")

        date = DateValue.parse("INT @#DGREGORIAN@ 1 JAN 2017 (some phrase)")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "INTERPRETED $date ($phrase)")
        self.assertEqual(date.kw, {"date": CalendarDate("2017", "JAN", 1),
                                   "phrase": "some phrase"})
        self.assertEqual(date.fmt(), "INTERPRETED 2017 JAN 1 (some phrase)")

    def test_016_date_parse_simple(self):
        """Test detail.date.DateValue class."""

        date = DateValue.parse("1967A.D.")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "$date")
        self.assertEqual(date.kw, {"date": CalendarDate("1967A.D.")})
        self.assertEqual(date.fmt(), "1967A.D.")

        date = DateValue.parse("@#DGREGORIAN@ 1 JAN 2017")
        self.assertTrue(date is not None)
        self.assertEqual(date.template, "$date")
        self.assertEqual(date.kw, {"date": CalendarDate("2017", "JAN", 1)})
        self.assertEqual(date.fmt(), "2017 JAN 1")

    def test_017_date_cmp(self):
        """Test detail.date.Date class."""

        dv = DateValue.parse("2016")
        self.assertIsInstance(dv._cmp_date, CalendarDate)
        self.assertEqual(dv._cmp_date, CalendarDate("2016"))

        dv = DateValue.parse("31 DEC 2000")
        self.assertIsInstance(dv._cmp_date, CalendarDate)
        self.assertEqual(dv._cmp_date, CalendarDate("2000", "DEC", 31))

        dv = DateValue.parse("BET 31 DEC 2000 AND 1 JAN 2001")
        self.assertIsInstance(dv._cmp_date, CalendarDate)
        self.assertEqual(dv._cmp_date, CalendarDate("2000", "DEC", 31))

        # earliest date
        dv = DateValue.parse("BET 31 DEC 2000 AND 1 JAN 2000")
        self.assertIsInstance(dv._cmp_date, CalendarDate)
        self.assertEqual(dv._cmp_date, CalendarDate("2000", "JAN", 1))

        self.assertTrue(DateValue.parse("2016") < DateValue.parse("2017"))
        self.assertTrue(DateValue.parse("2 JAN 2016") > DateValue.parse("1 JAN 2016"))
        self.assertTrue(DateValue.parse("BET 1900 AND 2000") < DateValue.parse("FROM 1920 TO 1999"))

        # Less specific date compares later than more specific
        self.assertTrue(DateValue.parse("2000") > DateValue.parse("31 DEC 2000"))
        self.assertTrue(DateValue.parse("DEC 2000") > DateValue.parse("31 DEC 2000"))

        # phrase is always later than any regular date
        self.assertTrue(DateValue.parse("(Could be 1996 or 1998)") > DateValue.parse("2000"))

        # "empty" date is always later than any regular date
        self.assertTrue(DateValue() > DateValue.parse("2000"))

    def test_018_date_parse_empty(self):
        """Test detail.date.DateValue class."""

        for value in (None, ""):
            date = DateValue.parse(value)
            self.assertTrue(date is not None)
            self.assertEqual(date.template, "")
            self.assertEqual(date.kw, {})
            self.assertEqual(date.fmt(), "")



"""Module containing methods for parsing GEDCOM files.
"""


from __future__ import print_function, absolute_import, division

__all__ = ['GedcomReader', 'ParserError', 'CodecError', 'IntegrityError',
           'guess_codec']

import codecs
import collections
import io
import logging
import re

from .detail.io import check_bom, guess_lineno
from . import model

_log = logging.getLogger(__name__)

# records are bytes, regex is for bytes too
_re_gedcom_line = re.compile(br"""
        ^
        [ ]*(?P<level>\d+)                       # integer level number
        (?:[ ]*(?P<xref>@[A-Z-a-z0-9][^@]*@))?    # optional @xref@
        [ ]*(?P<tag>[A-Z-a-z0-9_]+)               # tag name
        (?:[ ](?P<value>.*))?                    # optional value
        $
""", re.X)

# tuple class for gedcom_line grammar:
#   level: int
#   xref_id: str, possibly empty or None
#   tag: str, required, non-empty
#   value: bytes, possibly empty or None
#   offset: int
gedcom_line = collections.namedtuple("gedcom_line",
                                     "level xref_id tag value offset")


class ParserError(Exception):
    """Class for exceptions raised for parsing errors.
    """
    pass


class IntegrityError(Exception):
    """Class for exceptions raised for structural errors, e.g. when record
    level nesting is inconsistent.
    """
    pass


class CodecError(ParserError):
    """Class for exceptions raised for codec-related errors.
    """
    pass


def guess_codec(file, errors="strict", require_char=False, warn=True):
    """Look at file contents and guess its correct encoding.

    File must be open in binary mode and positioned at offset 0. If BOM
    record is present then it is assumed to be UTF-8 or UTF-16 encoded
    file. GEDCOM header is searched for CHAR record and encoding name
    is extracted from it, if BOM record is present then CHAR record
    must match BOM-defined encoding.

    :param file: File object, must be open in binary mode.
    :param str errors: Controls error handling behavior during string
        decoding, accepts same values as standard `codecs.decode` method.
    :param bool require_char: If True then exception is thrown if CHAR
        record is not found in a header, if False and CHAR is not in the
        header then codec determined from BOM or "gedcom" is returned.
    :param bool warn: If True (default) then generate error/warning messages
        for illegal encodings.
    :returns: Tuple (codec_name, bom_size)
    :raises: :py:class:`CodecError` when codec name in file is unknown or
        when codec name in file contradicts codec determined from BOM.
    :raises: :py:class:`UnicodeDecodeError` when codec fails to decode
        input lines and `errors` is set to "strict" (default).
    """

    # set of illegal but unambiguous encodings and their corresponding codecs
    illegal_encodings = {
        "windows-1250": "cp1250",
        "windows-1251": "cp1251",
        "cp1252": "cp1252",
        "iso-8859-1": "iso8859-1",
        "iso8859-1": "iso8859-1",
    }
    # set of ambiguous (and illegal) encodings
    ambiguous_encodings = {
        'ibmpc': 'cp437',
        "ibm": "cp437",
        "ibm-pc": "cp437",
        "oem": "cp437",
        "msdos": "cp850",
        "ibm dos": "cp850",
        "ms-dos": "cp850",
        "ansi": "cp1252",
        "windows": "cp1252",
        "ibm_windows": "cp1252",
        "ibm windows": "cp1252",
        "iso8859": "iso8859-1",
        "latin1": "iso8859-1",
        "macintosh": "mac-roman",
    }
    illegal_encodings.update(ambiguous_encodings)
    # full set of encodings, including legal ones
    gedcom_char_to_codec = {"ansel": "gedcom"}
    gedcom_char_to_codec.update(illegal_encodings)

    # check BOM first
    bom_codec = check_bom(file)
    bom_size = file.tell()
    codec = bom_codec or 'gedcom'

    # scan header until CHAR or end of header
    lineno = 0
    while True:

        lineno += 1

        # this stops at '\n'
        line = file.readline()
        if not line:
            raise IOError("Unexpected EOF while reading GEDCOM header")

        # do not decode bytes to strings here, reason is that some
        # stupid apps split CONC record at byte level (in middle of
        # of multi-byte characters). This implies that we can only
        # work with encodings that have ASCII as single-byte subset.

        line = line.lstrip().rstrip(b"\r\n")
        words = line.split()

        if len(words) >= 2 and words[0] == b"0" and words[1] != b"HEAD":
            # past header but have not seen CHAR
            if require_char:
                raise CodecError("GEDCOM header does not have CHAR record")
            else:
                break
        elif len(words) >= 3 and words[0] == b"1" and words[1] == b"CHAR":
            try:
                enc = b" ".join(words[2:]).decode(codec, errors)
                encoding = gedcom_char_to_codec.get(enc.lower(), enc.lower())
                if enc.lower() in illegal_encodings and warn:
                    _log.error("Line %d: \"%s\" - \"%s\" is not a legal "
                               "character set or encoding.", lineno, line, enc)
                    if enc.lower() in ambiguous_encodings:
                        _log.warning("Character set (\"%s\") is ambiguous, it "
                                     "will be interpreted as \"%s\"",
                                     enc, encoding)
                new_codec = codecs.lookup(encoding).name
            except LookupError:
                raise CodecError("Unknown codec name '{0}'".format(enc))
            if bom_codec is None:
                codec = new_codec
            elif new_codec != bom_codec:
                raise CodecError("CHAR codec {0} is different from BOM "
                                 "codec {1}".format(new_codec, bom_codec))
            break

    return codec, bom_size


class GedcomReader(object):
    """Main interface for reading GEDCOM files.

    :param file: File name or file object open in binary mode, file must
        be seekable.
    :param str encoding: If None (default) then file is analyzed using
        `guess_codec()` method to determine correct codec. Otherwise
        file is open using specified codec.
    :param str errors: Controls error handling behavior during string
        decoding, accepts same values as standard `codecs.decode` method.
    :param bool require_char: If True then exception is thrown if CHAR
        record is not found in a header, if False and CHAR is not in the
        header then codec determined from BOM or "gedcom" is used.
    """

    def __init__(self, file, encoding=None, errors="strict",
                 require_char=False):
        self._encoding = encoding
        self._errors = errors
        self._bom_size = 0
        self._index0 = None   # list of level=0 record positions
        self._xref0 = None    # maps xref_id to level=0 record position
        self._header = None
        self._dialect = None

        # open the file
        if hasattr(file, 'read'):
            # assume it is a file already
            if hasattr(file, 'seekable'):
                # check that it supports seek()
                if not file.seekable():
                    raise IOError("Input file does not support seek.")
            self._file = file
        else:
            raw = io.FileIO(file)
            self._file = io.BufferedReader(raw)

        # check codec and BOM
        try:
            encoding, self._bom_size = guess_codec(self._file,
                                                   errors=self._errors,
                                                   require_char=require_char,
                                                   warn=self._encoding is None)
        except Exception:
            self._file.close()
            raise
        self._file.seek(self._bom_size)
        if not self._encoding:
            self._encoding = encoding

    @property
    def index0(self):
        """List of level=0 record positions and tag names.
        """
        if self._index0 is None:
            self._init_index()
        return self._index0

    @property
    def xref0(self):
        """Dictionary which maps xref_id to level=0 record position and
        tag name.
        """
        if self._xref0 is None:
            self._init_index()
        return self._xref0

    @property
    def header(self):
        """Header record.
        """
        if self._index0 is None:
            self._init_index()
        return self._header

    def _init_index(self):
        _log.debug("in _init_index")
        self._index0 = []
        self._xref0 = {}
        # scan whole file for level=0 records
        for gline in self.gedcom_lines(self._bom_size):
            _log.debug("  _init_index gline: %s", gline)
            if gline.level == 0:
                self._index0.append((gline.offset, gline.tag))
                if gline.xref_id:
                    self._xref0[gline.xref_id] = (gline.offset, gline.tag)
            _log.debug("  _init_index gline: done proc")
        if self._index0 and self._index0[0][1] == 'HEAD':
            self._header = self.read_record(self._index0[0][0])
        _log.debug("_init_index done")

    @property
    def dialect(self):
        """File dialect as one of model.DIALECT_* constants
        """
        if self._dialect is None:
            self._dialect = model.DIALECT_DEFAULT
            if self.header:
                source = self.header.sub_tag("SOUR")
                if source:
                    if source.value == "MYHERITAGE":
                        self._dialect = model.DIALECT_MYHERITAGE
                    elif source.value == "ALTREE":
                        self._dialect = model.DIALECT_ALTREE
                    elif source.value == "ANCESTRIS":
                        self._dialect = model.DIALECT_ANCESTRIS
        return self._dialect

    @dialect.setter
    def dialect(self, value):
        self._dialect = value

    def gedcom_lines(self, offset):
        """Generator method for *gedcom lines*.

        GEDCOM line grammar is defined in Chapter 1 of GEDCOM standard, it
        consists of the level number, optional reference ID, tag name, and
        optional value separated by spaces. Chaper 1 is pure grammar level,
        it does not assign any semantics to tags or levels. Consequently
        this method does not perform any operations on the lines other than
        returning the lines in their order in file.

        This method iterates over all lines in input file and converts each
        line into :py:class:`gedcom_line` class.

        :param int offset: Position in the file to start reading.
        :returns: Iterator for gedcom_lines.
        :raises: :py:class:`ParserError` when lines have incorrect syntax.
        """

        self._file.seek(offset)

        prev_gline = None
        while True:

            offset = self._file.tell()
            line = self._file.readline()  # stops at \n
            if not line:
                break
            line = line.lstrip().rstrip(b"\r\n")

            match = _re_gedcom_line.match(line)
            if not match:
                self._file.seek(offset)
                lineno = guess_lineno(self._file)
                line = line.decode(self._encoding, "ignore")
                raise ParserError("Invalid syntax at line "
                                  "{0}: `{1}'".format(lineno, line))

            level = int(match.group('level'))
            xref_id = match.group('xref')
            if xref_id:
                xref_id = xref_id.decode(self._encoding, self._errors)
            tag = match.group('tag').decode(self._encoding, self._errors)

            # simple structural integrity check
            if prev_gline is not None:
                if level - prev_gline.level > 1:
                    # nested levels should be incremental (+1)
                    self._file.seek(offset)
                    lineno = guess_lineno(self._file)
                    line = line.decode(self._encoding, "ignore")
                    raise IntegrityError("Structural integrity - "
                                         "illegal level nesting at line "
                                         "{0}: `{1}'".format(lineno, line))
                if tag in ("CONT", "CONC"):
                    # CONT/CONC level must be +1 from preceding non-CONT/CONC
                    # record or the same as preceding CONT/CONC record
                    if ((prev_gline.tag in ("CONT", "CONC") and
                         level != prev_gline.level) or
                        (prev_gline.tag not in ("CONT", "CONC") and
                         level - prev_gline.level != 1)):
                        self._file.seek(offset)
                        lineno = guess_lineno(self._file)
                        line = line.decode(self._encoding, "ignore")
                        raise IntegrityError("Structural integrity -  illegal "
                                             "CONC/CONT nesting at line "
                                             "{0}: `{1}'".format(lineno, line))

            gline = gedcom_line(level=level,
                                xref_id=xref_id,
                                tag=tag,
                                value=match.group('value'),
                                offset=offset)
            yield gline

            prev_gline = gline

    def records0(self, tag=None):
        """Iterator over all level=0 records.

        :param str tag: If ``None`` is given (default) then return all level=0
            records, otherwise return level=0 records with the given tag.
        """
        _log.debug("in records0")
        for offset, xtag in self.index0:
            _log.debug("    records0: offset: %s; xtag: %s", offset, xtag)
            if tag is None or tag == xtag:
                yield self.read_record(offset)

    def read_record(self, offset):
        """Read next complete record from a file starting at given position.

        Reads the record at given position and all its sub-records. Stops
        reading at EOF or next record with the same or higher (smaller) level
        number. File position after return from this method is not specified,
        re-position file if you want to read other records.

        :param int offset: Position in file to start reading from.
        :return: :py:class:`model.Record` instance or None if offset points
            past EOF.
        :raises: :py:exc:`ParserError` if `offsets` does not point to the
            beginning of a record or for any parsing errors.
        """
        _log.debug("in read_record(%s)", offset)
        stack = []  # stores per-level current records
        reclevel = None
        for gline in self.gedcom_lines(offset):
            _log.debug("    read_record, gline: %s", gline)
            level = gline.level

            if reclevel is None:
                # this is the first record, remember its level
                reclevel = level
            elif level <= reclevel:
                # stop at the record of the same or higher (smaller) level
                break

            # All previously seen records at this level and below can
            # be finalized now
            for rec in reversed(stack[level:]):
                # decode bytes value into string
                if rec:
                    if rec.value is not None:
                        rec.value = rec.value.decode(self._encoding,
                                                     self._errors)
                    rec.freeze()
#                    _log.debug("    read_record, rec: %s", rec)
            del stack[level + 1:]

            # extend stack to fit this level (and make parent levels if needed)
            stack.extend([None] * (level + 1 - len(stack)))

            # make Record out of it (it can be updated later)
            parent = stack[level - 1] if level > 0 else None
            rec = self._make_record(parent, gline)

            # store as current record at this level
            stack[level] = rec

        for rec in reversed(stack[reclevel:]):
            if rec:
                if rec.value is not None:
                    rec.value = rec.value.decode(self._encoding, self._errors)
                rec.freeze()
                _log.debug("    read_record, rec: %s", rec)

        return stack[reclevel] if stack else None

    def _make_record(self, parent, gline):
        """Process next record.

        This method created new record from the line read from file if
        needed and/or updates its parent record. If the parent record tag
        is ``BLOB`` and new record tag is ``CONT`` then record is skipped
        entirely and None is returned. Otherwise if new record tag is ``CONT``
        or ``CONC`` its value is added to parent value. For all other tags
        new record is made and it is added to parent sub_records attribute.

        Parameters
        ----------
        parent : `model.Record`
            Parent record of the new record
        gline : `gedcom_line`
            Current parsed line

        Returns
        -------
        `model.Record` or None
        """

        if parent and gline.tag in ("CONT", "CONC"):
            # concatenate, only for non-BLOBs
            if parent.tag != "BLOB":
                # have to be careful concatenating empty/None values
                value = gline.value
                if gline.tag == "CONT":
                    value = b"\n" + (value or b"")
                if value is not None:
                    parent.value = (parent.value or b"") + value
            return None

        # avoid infinite cycle
        dialect = model.DIALECT_DEFAULT
        if not (gline.level == 0 and gline.tag == "HEAD") and self._header:
            dialect = self.dialect
        rec = model.make_record(level=gline.level, xref_id=gline.xref_id,
                                tag=gline.tag, value=gline.value,
                                sub_records=[], offset=gline.offset,
                                dialect=dialect, parser=self)

        # add to parent's sub-records list
        if parent:
            parent.sub_records.append(rec)

        return rec

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._file.close()
